from __future__ import annotations

from dataclasses import dataclass, field
from itertools import islice, pairwise
from typing import Iterable, Any, Generator, cast

import numpy
import numpy.typing as npt
from scipy.ndimage import gaussian_filter1d
from scipy.signal import find_peaks as scipy_signal_find_peaks
from scipy.optimize import minimize_scalar
from skimage.filters import threshold_otsu

try:
    # python 3.12?
    from itertools import batched  # type: ignore[attr-defined]
except ImportError:
    def batched(iterable: Iterable[Any], n: int) -> Iterable[tuple[Any, ...]]:
        # batched('ABCDEFG', 3) --> ABC DEF G
        if n < 1:
            raise ValueError('n must be at least one')
        it = iter(iterable)
        while batch := tuple(islice(it, n)):
            yield batch


def pad_to_signal(pulse_and_durations: npt.NDArray[numpy.int64]) -> npt.NDArray[numpy.int8]:
    """
    Convert pulse and duration values from flipper to a binary (0/1) signal

    :param pulse_and_durations: The pulse and duration values
    :return: reconstructed signal
    """
    length = pulse_and_durations[:, 1].sum()
    signal = numpy.zeros(length, dtype=numpy.int8)

    position = 0
    for pulse, duration in pulse_and_durations:
        # Fill signal with ones for pulse
        signal[position:position + pulse] = 1
        # Fill signal with zeros for the rest of the duration (not needed, because signal default to zero, just for clarity)
        # signal[position+pulse:position+duration] = 0
        position += duration

    return signal


def signal_to_pad(signal: npt.NDArray[numpy.int8]) -> npt.NDArray[numpy.int64]:
    """
    Convert a binary (0/1) signal to pulse and duration values like used in flipper

    :param signal: The signal
    :return: Pulse and duration values
    """
    def it(s: npt.NDArray[numpy.int8]) -> Generator[tuple[int, int], None, None]:
        position = -1
        changes = numpy.where(s[:-1] != s[1:])[0]
        for p in batched(changes, 2):
            if len(p) == 2:
                yield p[0] - position, p[1] - position
                position = p[1]

        if len(p) == 1:
            yield p[0] - position, len(s) - position - 1

    return numpy.array(list(it(signal)))


def find_first_transition_index(signal: npt.NDArray[numpy.int8], to: int = 1) -> int:
    """
    Find the first index in the signal where it transitions to `to

    :param signal: the signal
    :param to: the value to transition to
    :return: index of the first transition
    """
    # Array of the first 2 indices where there is a transition in any direction
    changes = (signal[:-1] != signal[1:]).nonzero()[0][:2] + 1
    # One of them is the change to the requested `to` value
    needed_index = (signal[changes] == to).nonzero()[0][0]

    change_index = changes[needed_index]
    # Some clarity after too much numpy magic ;)
    assert signal[change_index] == to
    assert signal[change_index - 1] != to

    return cast(int, change_index)


@dataclass
class Peak:
    """
    A peak in a distribution described by left, center and right index
    """
    left: int = field(compare=False)
    center: int = field(compare=False)
    right: int = field(compare=False)
    height: float = field(default=0.0, repr=False)

    def merge(self, other: Peak) -> Peak:
        """
        Merge this peak with another peak
        :param other: Peak to merge with
        :return: merged peak
        """
        return Peak(
            min(self.left, other.left),
            (self.center + other.center) // 2,
            max(self.right, other.right),
            max(self.height, other.height)
        )

    def slice(self, distribution: npt.NDArray[Any]) -> npt.NDArray[Any]:
        """
        Slice the distribution with the peak

        :param distribution:
        :return:
        """
        return distribution[self.left:self.right]

    def fit(self, distribution: npt.NDArray[Any], quantile: float = 1.0) -> Peak:
        """
        Fit the distribution to the peak
        :param distribution:
        :param quantile:
        :return:
        """
        my_excerpt = distribution[self.left:self.right]
        if quantile < 1.0:
            to_capture = numpy.sum(my_excerpt) * quantile

            def objective(thr: float) -> float:
                # 1.0 for capturing enough and a little nudge to find bigger thresholds
                return cast(float, 1.0 * (to_capture > numpy.sum(my_excerpt[my_excerpt > thr])) - thr * 0.0001)

            res = minimize_scalar(objective, (0, my_excerpt.max()))
            threshold = int(res.x)
        else:
            threshold = 0

        first, *_, last = (my_excerpt > threshold).nonzero()[0]

        return Peak(
            self.left + first - 1,
            self.left + (first + last) // 2,
            self.left + last + 1,
            my_excerpt[first:last].max()
        )

    def __contains__(self, v: float | int) -> bool:
        """
        Check if a value is inside the peak
        :param v: value to check
        :return:
        """
        return self.left <= v <= self.right


def histogram(values: npt.NDArray[Any], min_length: int = None) -> npt.NDArray[numpy.int32]:
    """
    Calculate a stupid histogram of a distribution, each value is it's own "bin"

    :param values: The values to count
    :param min_length: Optional minimum length of histogram
    :return: histogram
    """
    length = max(values.max() + 20, min_length or 0)
    hist = numpy.zeros(length, dtype=numpy.int32)
    for v in values:
        hist[v] += 1
    return hist


def find_peaks(distribution: npt.NDArray[Any], min_height: float = None, separate_peaks: bool = True) -> list[Peak]:
    """
    Simple wrapper around scipy.signal.find_peaks auto-tuned for histograms and sorted Peak[] as return value

    :param distribution: The signal to analyze
    :param min_height: optional minimum height for peaks, if None, mean of distribution is used
    :param separate_peaks:  if True, separate peaks that have overlap via otsu thresholding
    :return: array of Peak
    """
    if min_height is None:
        min_height = numpy.mean(distribution)
    peaks_center, pd = scipy_signal_find_peaks(distribution, height=min_height, prominence=min_height)
    peaks = [Peak(l, c, r, h) for c, l, r, h in zip(peaks_center, pd['left_bases'], pd['right_bases'], pd['peak_heights'])]

    # Separate peaks that have overlap
    if separate_peaks:
        for left, right in pairwise(sorted(peaks, key=lambda p: p.center)):
            if left.right > right.left:
                left.right = right.left = left.left + threshold_otsu(hist=distribution[left.left:right.right])

    return sorted(peaks, key=lambda p: p.height, reverse=True)


def autocorrelate(x: npt.NDArray[Any]) -> npt.NDArray[numpy.float32]:
    """
    Calculate fast statistical autocorrelation

    :param x: signal
    :return: autocorrelation of signal

    taken from:
    https://stackoverflow.com/a/51168178
    autocorr3 / fft, pad 0s, non partial
    """
    n = len(x)
    # pad 0s to 2n-1
    ext_size = 2 * n - 1
    # nearest power of 2
    fsize = 2 ** numpy.ceil(numpy.log2(ext_size)).astype('int')

    xp = x - numpy.mean(x)
    var = numpy.var(x)

    # do fft and ifft
    cf = numpy.fft.fft(xp, fsize)
    sf = cf.conjugate() * cf
    corr = numpy.fft.ifft(sf).real
    corr = corr / var / n

    return (corr[:len(corr) // 2]).astype(numpy.float32)


def smooth(signal: npt.NDArray[numpy.int8], sigma: float = 10) -> npt.NDArray[numpy.float32]:
    """
    Apply gaussian filtering to signal

    :param signal: The input signal
    :param sigma: sigma for gaussian filter, how much smoothing
    :return: smoothed signal
    """
    return cast(npt.NDArray[numpy.float32], gaussian_filter1d(numpy.float32(signal), sigma=sigma, mode='nearest'))


def binarize(signal: npt.NDArray[numpy.float32], threshold: float = 0.5) -> npt.NDArray[numpy.int8]:
    """
    Binarize (0/1) signal with threshold

    :param signal: The input signal
    :param threshold: threshold for binarization
    :return: binarized signal
    """
    return (signal > threshold).astype(numpy.int8)


__all__ = ['signal_to_pad', 'pad_to_signal', 'smooth', 'binarize', 'autocorrelate', 'batched', 'Peak', 'find_first_transition_index', 'find_peaks', 'histogram']



from itertools import zip_longest
from pathlib import Path
from unittest import TestCase

import numpy
from numpy.testing import assert_array_equal

from flipper_raw_rfid.rifl import Rifl
from flipper_raw_rfid.utils import pad_to_signal, signal_to_pad

TEST_BASE_PATH = (Path(__file__).parent).absolute()


class UtilsTest(TestCase):

    @property
    def a_signal(self):
        signal = numpy.zeros(10000, dtype=numpy.int8)
        start = 0
        signal[start:start + 310] = 1
        start += 515
        signal[start:start + 274] = 1
        start += 527
        signal[start:start + 252] = 1
        start += 743
        signal[start:start + 291] = 1
        start += 534
        signal[start:start + 515] = 1
        start += 1016
        signal[start:start + 266] = 1
        start += 515

        return signal[:start]

    @property
    def a_pad(self):
        return numpy.array([
            [310, 515],
            [274, 527],
            [252, 743],
            [291, 534],
            [515, 1016],
            [266, 515]
        ])

    def load(self, file: str | Path) -> Rifl:
        return Rifl.load(TEST_BASE_PATH / 'assets' / file)

    def test_pad_to_signal(self):

        signal = pad_to_signal(self.a_pad)

        # First row, a pulse from 0-309 then a zero from 310 - 514
        self.assertEqual(signal[0], 1)
        self.assertEqual(signal[309], 1)
        self.assertEqual(signal[310], 0)
        self.assertEqual(signal[514], 0)

        # Second row, a pulse for 274 samples starting at 515
        self.assertEqual(signal[515], 1)
        self.assertEqual(signal[515 + 273], 1)
        self.assertEqual(signal[515 + 274], 0)
        self.assertEqual(signal[515 + 527 - 1], 0)

        self.assertEqual(signal[515 + 527], 1)
        # and so on

        assert_array_equal(signal, self.a_signal)

    def test_signal_to_pad(self):
        pad = signal_to_pad(self.a_signal)
        assert_array_equal(pad, self.a_pad)

    def test_signal_to_pad_and_back(self):

        def test(signal):
            rec_signal = pad_to_signal(signal_to_pad(signal))
            assert_array_equal(signal, rec_signal)

        test(self.a_signal)
        test(pad_to_signal(self.load('Red354b.ask.raw').pulse_and_durations))

    def test_pad_to_signal_and_back(self):

        def test(pad):
            rec_pad = signal_to_pad(pad_to_signal(pad))

            pos = 0
            for i, (pd, rpd) in enumerate(zip_longest(pad, rec_pad)):
                if not numpy.array_equal(pd, rpd):
                    assert_array_equal(pd, rpd, err_msg=f'Difference in reconstructed pad in row {i}, sample {pos}')
                pos += pd[1]

        test(self.a_pad)
        test(self.load('Red354b.ask.raw').pulse_and_durations)
        # This fails because of 0 pulse width
        # test(self.load_pad('Red354.ask.raw'))



"""
Classes to load/save a raw rfid file from flipper (xyz.ask.raw or xyz.psk.raw)

Usage:

rifl = RiflFile.load('path/to/raw.ask.raw')
# get header values
frequency = rifl.header.frequency
duty_cycle = rifl.header.duty_cycle
# get pulse and duration values
pd = rifl.pulse_and_durations

"""
from __future__ import annotations

from io import BytesIO
from typing import BinaryIO, Generator, Any
from struct import unpack, pack, error as struct_error
from pathlib import Path
from dataclasses import dataclass

import numpy
import numpy.typing as npt

from flipper_raw_rfid.utils import batched

LFRFID_RAW_FILE_MAGIC = 0x4C464952  # binary string "RIFL"
LFRFID_RAW_FILE_VERSION = 1


class RiflError(ValueError):
    def __init__(self, message: Any, file: BinaryIO = None):
        super().__init__(message)
        self.file = file


@dataclass
class RiflHeader:
    """
    Rifl Header data structure
    """
    version: int
    """ Version of the rifl file format: 1 supported """
    frequency: float
    """ Frequency of the signal in Hz """
    duty_cycle: float
    """ Duty cycle of the signal"""
    max_buffer_size: int
    """ Maximum buffer size in bytes"""

    @staticmethod
    def from_io(io: BinaryIO) -> RiflHeader:
        try:
            return RiflHeader.from_bytes(io.read(20))
        except RiflError as e:
            e.file = io
            raise e

    @staticmethod
    def from_bytes(f: bytes) -> RiflHeader:
        try:
            magic, version, frequency, duty_cycle, max_buffer_size = unpack('IIffI', f)
        except struct_error:
            raise RiflError('Not a RIFL file')
        if magic != LFRFID_RAW_FILE_MAGIC:
            raise RiflError('Not a RIFL file')
        if version != LFRFID_RAW_FILE_VERSION:
            raise RiflError(f'Unsupported RIFL Version {version}')

        return RiflHeader(version, frequency, duty_cycle, max_buffer_size)

    def to_bytes(self) -> bytes:
        return pack('IIffI', LFRFID_RAW_FILE_MAGIC, self.version, self.frequency, self.duty_cycle, self.max_buffer_size)


@dataclass
class Rifl:
    """
    A raw rfid file from flipper (xyz.ask.raw or xyz.psk.raw)

    """
    header: RiflHeader
    """ The header of the file """

    pulse_and_durations: npt.NDArray[numpy.int64] = None
    """
    a nx2 numpy array with:
    column 0: pulse - (number of µs while output high) and
    column 1: duration - (number of µs till next signal)

    Diagram:

    _____________      _____
                 ______     _______________ .......

    ^ - pulse - ^

    ^ -    duration  -^


    """

    @staticmethod
    def load(path: Path | str) -> Rifl:
        path = Path(path)
        with path.open('rb') as f:
            return Rifl.from_io(f)

    @staticmethod
    def from_io(io: BinaryIO) -> Rifl:
        header = RiflHeader.from_io(io)
        pads = numpy.array(list(Rifl._pulse_and_durations(io, header.max_buffer_size)), dtype=numpy.int64)
        return Rifl(header, pads)

    def save(self, path: Path | str) -> None:
        path = Path(path)
        with path.open('wb') as f:
            self.to_io(f)

    def to_io(self, io: BinaryIO) -> None:

        def write(b: BytesIO) -> None:
            io.write(pack('I', b.getbuffer().nbytes))
            io.write(b.getvalue())

        def write_pair(b: BytesIO, pair: BytesIO) -> BytesIO:
            if b.getbuffer().nbytes + pair.getbuffer().nbytes > self.header.max_buffer_size:
                write(b)
                b = BytesIO()
            b.write(pair.getvalue())
            return b

        io.write(self.header.to_bytes())

        buffer = BytesIO()
        for pulse, duration in self.pulse_and_durations:
            pair_buffer = BytesIO()
            Rifl.write_varint(pair_buffer, pulse)
            Rifl.write_varint(pair_buffer, duration)
            buffer = write_pair(buffer, pair_buffer)

        write(buffer)

    @staticmethod
    def _buffers(io: BinaryIO, max_buffer_size: int) -> Generator[BinaryIO, None, None]:
        """
        Read raw binary buffers  and loop through them

        Each buffer holds varint (https://github.com/flipperdevices/flipperzero-firmware/blob/dev/lib/toolbox/varint.c#L13) encoded pairs
        """
        while True:
            try:
                buffer_size, = unpack('I', io.read(4))
            except struct_error:
                # No more bytes left, EOF
                break
            if buffer_size > max_buffer_size:
                raise RiflError(f'read pair: buffer size is too big  {buffer_size} > {max_buffer_size}', io)
            buffer = io.read(buffer_size)
            if len(buffer) != buffer_size:
                raise RiflError(f'Tried to read  {buffer_size} bytes got only {len(buffer)}', io)
            yield BytesIO(buffer)

    @staticmethod
    def _pulse_and_durations(io: BinaryIO, max_buffer_size: int) -> Generator[tuple[int, int], None, None]:
        """
        loop through buffers and yield a pulse and duration tuple
        """
        for buffer in Rifl._buffers(io, max_buffer_size):
            for pulse, duration in batched(Rifl.read_varint(buffer), 2):
                yield pulse, duration

    @staticmethod
    def read_varint(buffer: BinaryIO) -> Generator[int, None, None]:
        """
        Read one varint from buffer

        Python implementation of https://github.com/flipperdevices/flipperzero-firmware/blob/dev/lib/toolbox/varint.c#L13

        """
        res = 0
        i = 1
        while (vs := buffer.read(1)) != b'':
            v = vs[0]
            # the low 7 bits are the value
            res = res | (v & 0x7F) * i
            i = i << 7
            # yield when continue bit (bit 8) is not set
            if v & 0x80 == 0:
                yield res
                res = 0
                i = 1

    @staticmethod
    def write_varint(buffer: BinaryIO, value: int) -> int:
        """
        Write one varint to buffer
        """
        i = 1
        while value > 0x80:
            buffer.write(bytes([value & 0x7F | 0x80]))
            value >>= 7
            i += 1

        buffer.write(bytes([value & 0x7F]))
        return i


__all__ = ['Rifl', 'RiflError']



"""
Utilities for working with bitstreams
"""
import re
import numpy
from flipper_raw_rfid.utils import batched, Peak
import numpy.typing as npt


def decode_lengths(pads: npt.NDArray[numpy.int64], peaks: list[Peak]) -> tuple[npt.NDArray[numpy.int8], int]:
    """
    Loops through pulses and durations and matches them to peaks
    Checks for the length of the peak as a multiple of the first peak and adds as many 1/0 to the result

    :param pads: Pulse and duration values
    :param peaks: A list of peaks from find_peaks, the center frequencies should be more or less multiples of the first peak
    :return: The decoded bitstream
    """
    result: list[int] = []
    position = 0
    result_position = None
    first_length = peaks[0].center
    for high, duration in pads:
        low = duration - high

        high_peak = None
        low_peak = None

        for p in peaks:
            if high in p:
                high_peak = p
            if low in p:
                low_peak = p
            if high_peak and low_peak:
                break

        if not (high_peak and low_peak):
            if not high_peak:
                print(f'Found nothing for high {high}, restarting')
            if not low_peak:
                print(f'Found nothing for low {low}, restarting')
            result = []
            result_position = position
            continue

        result.extend([1] * int(round(high_peak.center / first_length)))
        result.extend([0] * int(round(low_peak.center / first_length)))
        position += duration

    return numpy.array(result, dtype=numpy.int8), result_position


def decode_manchester(manchester: npt.NDArray[numpy.int8], biphase: bool = True) -> npt.NDArray[numpy.int8]:
    """
    Decode manchester encoded bitstream
    :param manchester: manchester encoded bitstream
    :param biphase: True for biphase, False for diphase
    :return: decoded bitstream
    """
    if manchester[0] == manchester[1]:
        manchester = manchester[1:]

    result = []
    for pair in batched(manchester, 2):
        if len(pair) < 2:
            break
        assert pair[0] != pair[1]
        result.append(pair[0 if biphase else 1])

    return numpy.array(result, dtype=numpy.int8)


def to_str(bits: npt.NDArray[numpy.int8]) -> str:
    """
    Convert a bitstream to a string
    :param bits:
    :return:
    """
    return ''.join(str(b) for b in bits)


def find_pattern(bits: npt.NDArray[numpy.int8], pattern: str | re.Pattern[str]) -> npt.NDArray[numpy.int8] | None:
    bitstring = ''.join(str(b) for b in bits)
    m = re.search(pattern, bitstring)
    if not m:
        return None

    return bits[m.start(0):m.end(0)]


def decode_em_4100(bits: npt.NDArray[numpy.int8]) -> npt.NDArray[numpy.int8]:
    """
    Decode bitstream as EM 4100

    :param bits: bitstream
    :return: decoded nibbles


    EM 4100
    has a header of 9 '1's
    111111111

    followed by 10 nibbles (4 bits each) of data plus one parity bit
    1010 0
    1000 1
    ...

    followed by 4 column parity bits and a final 0

    0010 0

    """

    em4100_bits = find_pattern(bits, r'1{9}.{54}0')

    datagrid = em4100_bits[9:].reshape((11, 5))

    column_parity = datagrid[:, :4].sum(axis=0)
    assert numpy.all(column_parity % 2 == 0)
    row_parity = datagrid[:10].sum(axis=1)
    assert numpy.all(row_parity % 2 == 0)
    nibbles = (datagrid[:10, :4] * [8, 4, 2, 1]).sum(axis=1)

    return numpy.array(nibbles, dtype=numpy.int8)


def longest_run(bits: npt.NDArray[numpy.int8], value: int = 1) -> int:
    """
    Find the longest run of a value (1/0) in a bitstream

    :param bits: the bitstream
    :param value: the value to look for
    :return: index of the first bit of the longest run
    """
    longest = 0
    current = 0
    longest_i = None

    for i, b in enumerate(bits):
        if b != value:
            current = 0
            continue
        current += 1
        if current > longest:
            longest = current
            longest_i = i

    return longest_i - longest + 1



#! /usr/bin/python3
"""flipper-raw-rfid

Description:
    Reads a raw rfid file from flipper zero and plots or converts the signal

Usage:
    flipper-raw-rfid convert [-f <format>] RAW_FILE [OUTPUT_FILE]
    flipper-raw-rfid plot RAW_FILE
    flipper-raw-rfid check RAW_FILE
    flipper-raw-rfid (-h | --help)
    flipper-raw-rfid --version

Arguments:
    RAW_FILE        The raw rfid file from flipper (xyz.ask.raw or xyz.psk.raw)
    OUTPUT_FILE     The converted file as csv (default: stdout)

Options:
    -h --help                 Show this screen.
    --version                 Show version.
    -f --format=(pad|signal)  Output format: "pad" (=Pulse And Duration) is the internal format of the Flipper Zero,
                              each line represents a pulse and a duration value measured in µs, see
                              "Pulse and duration format" below.
                              In "signal" format the pulses are written out as a reconstructed signal with a "1" marking a
                              sample with high value and "0" marking a sample with low value [default: pad]

Pulse and duration format:

    column 0: pulse - (µs while output high) and
    column 1: duration - (µs till next signal)

    Diagram:

    ______________      __________
                  ______          __________ .......

    ^ - pulse0 - ^      ^-pulse1-^           .......
    ^ -    duration0  -^^ -    duration1  -^ .......

    The csv file has the following format:

    pulse0, duration0
    pulse1, duration1
    ....

"""
import contextlib
import csv
import os
import sys
from typing import Any, Generator, Iterable, IO

from docopt import docopt, printable_usage
from flipper_raw_rfid import __version__
from flipper_raw_rfid.rifl import Rifl, RiflError
from flipper_raw_rfid.utils import pad_to_signal, autocorrelate
from scipy import signal as scipy_signal


@contextlib.contextmanager
def smart_open(filename: str, mode: str = 'r', *args: Any, **kwargs: Any) -> Generator[IO[Any], None, None]:
    """
    Open files and i/o streams transparently.
    """
    fh: IO[Any]
    if filename == '-' or filename is None:
        if 'r' in mode:
            stream = sys.stdin
        else:
            stream = sys.stdout
        if 'b' in mode:
            fh = stream.buffer
        else:
            fh = stream
        close = False
    else:
        fh = open(filename, mode, *args, **kwargs)
        close = True

    try:
        yield fh
        fh.flush()
    except BrokenPipeError:
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, fh.fileno())
    finally:
        if close:
            try:
                fh.close()
            except AttributeError:
                pass


def assert_in(value: Any, vset: Iterable[Any], name: str = 'Value') -> None:
    if value not in vset:
        raise ValueError(f'{name} must be one of: {"/".join(vset)} but was "{value}"')


def convert(raw: str, output: str, format: str = 'pad') -> int:
    assert_in(format, {'signal', 'pad'}, 'Format')
    with smart_open(output, 'w', newline='') as csvfile:
        csvwriter = csv.writer(csvfile)
        rifl = Rifl.load(raw)
        pads = rifl.pulse_and_durations
        if format == 'pad':
            for pulse_and_duration in pads:
                csvwriter.writerow(pulse_and_duration)
        elif format == 'signal':
            signal = pad_to_signal(pads)
            csvwriter.writerows(signal.reshape((-1, 1)))

    return 0


def plot(raw: str) -> int:
    try:
        import matplotlib.pyplot as plt
    except ModuleNotFoundError:
        raise RuntimeError('For the plot command you need matplotlib, install it with: pip install matplotlib')
    rifl = Rifl.load(raw)
    pads = rifl.pulse_and_durations
    signal = pad_to_signal(pads)
    fig, ax = plt.subplots()
    ax.set(xlabel='Samples', ylabel='Signal', title='Reconstructed Signal')
    ax.set_xlim(0, 20000)
    ax.plot(signal)
    plt.show()

    return 0


def check(raw: str) -> int:
    rifl = Rifl.load(raw)
    pads = rifl.pulse_and_durations
    signal = pad_to_signal(pads)
    signal_autocorrelation = autocorrelate(signal)
    peaks, _ = scipy_signal.find_peaks(signal_autocorrelation, height=0.8, prominence=1)
    print(peaks)
    return 0


def process(args: dict[str, Any]) -> int:
    try:
        if args['convert']:
            return convert(args['RAW_FILE'], args['OUTPUT_FILE'], args['--format'])
        elif args['plot']:
            return plot(args['RAW_FILE'])
        elif args['check']:
            return check(args['RAW_FILE'])
        else:
            print(printable_usage(__doc__))
            return 1
    except RiflError as e:
        print(printable_usage(__doc__))
        print(f'Error: {e}: {e.file.name if e.file.name else ""}')
        return 1
    except (ValueError, FileNotFoundError, IsADirectoryError, RuntimeError) as e:
        print(printable_usage(__doc__))
        print(f'Error: {e}')
        return 1


def main() -> int:
    args = docopt(__doc__, version=f'flipper-raw-rfid {__version__}')
    return process(args)



from io import BytesIO
from pathlib import Path
from unittest import TestCase

import numpy
from numpy.testing import assert_array_equal

from flipper_raw_rfid.rifl import Rifl, RiflHeader

TEST_BASE_PATH = Path(__file__).parent.absolute()


class RiflFileTest(TestCase):

    example_bytes = bytes.fromhex('f101a903ae028506a604fb05bb028706ad04b90404c403')
    example_ints = [241, 425, 302, 773, 550, 763, 315, 775, 557, 569, 4, 452]

    def test_header_to_bytes_and_back(self):
        header = RiflHeader(1, 125_000, 0.5, 2048)
        self.assertEqual(header, RiflHeader.from_bytes(header.to_bytes()))

    def test_header_checks_magic(self):
        RiflHeader.from_bytes(bytes.fromhex('5249464C 01000000 0024F447 0000003F  00080000'))

        with self.assertRaisesRegex(ValueError, 'Not a RIFL file'):
            RiflHeader.from_bytes(bytes.fromhex('C0FFEEAA 01000000 0024F447 0000003F  00080000'))

    def test_header_checks_version(self):
        header = RiflHeader(2, 125_000, 0.5, 2048)
        with self.assertRaisesRegex(ValueError, 'Unsupported RIFL Version 2'):
            RiflHeader.from_bytes(header.to_bytes())

    def test_read_varint(self):
        buffer = BytesIO(self.example_bytes)
        self.assertEqual(self.example_ints, list(Rifl.read_varint(buffer)))

    def test_write_varint(self):
        buffer = BytesIO()
        for v in self.example_ints:
            Rifl.write_varint(buffer, v)
        self.assertEqual(self.example_bytes, buffer.getvalue())

    def test_to_and_from_io(self):
        # Load file
        rifl = Rifl.load(TEST_BASE_PATH / 'assets' / 'Red354.ask.raw')
        buffer = BytesIO()
        rifl.to_io(buffer)

        # Load from buffer, should be the same
        rifl2 = Rifl.from_io(BytesIO(buffer.getvalue()))
        buffer2 = BytesIO()
        rifl2.to_io(buffer2)

        self.assertEqual(rifl.header, rifl2.header)
        assert_array_equal(rifl.pulse_and_durations, rifl2.pulse_and_durations)
        self.assertEqual(buffer.getvalue(), buffer2.getvalue())

    def test_save_and_load(self):
        dummy_file = TEST_BASE_PATH / 'assets' / 'dummy.ask.raw'

        header = RiflHeader(1, 125000.0, 0.5, 2048)
        pads = numpy.reshape(self.example_ints, (-1, 2))
        dummy1 = Rifl(header, pads)
        dummy1.save(dummy_file)

        dummy2 = Rifl.load(dummy_file)
        self.assertEqual(dummy1.header, dummy2.header)
        assert_array_equal(dummy1.pulse_and_durations, dummy2.pulse_and_durations)

        dummy_file.unlink()



"""
Some useful functions not used in the main code yet
"""
from typing import Any, Generator

import numpy
import numpy.typing as npt

from flipper_raw_rfid.utils import find_peaks


def correlation_offset(d1: npt.NDArray[numpy.int32], d2: npt.NDArray[numpy.int32]) -> int:
    """
    Find the offset where two signals would correlate best
    :param d1:
    :param d2:
    :return:
    """
    middle = len(d1) // 2
    cor = numpy.correlate(d1, d2, mode='same')
    return find_peaks(cor)[0].center - middle


def roll(a: npt.NDArray[Any], shift: int) -> npt.NDArray[Any]:
    """
    Like numpy roll but with padding instead of wrapping around
    :param a: array
    :param shift: amount to shift
    :return: shifted array
    """
    if shift > 0:
        return numpy.pad(a[:-shift], (shift, 0))
    else:
        return numpy.pad(a[-shift:], (0, -shift))


def rationalizations(x: float, rtol: float = 1e-05, atol: float = 1e-08) -> Generator[tuple[int, int], None, None]:
    """
    Find short rational approximations of a float

    :param x: the float to approximate
    :param rtol: Relative tolerance, see numpy.isclose
    :param atol: Absolute tolerance, see numpy.isclose
    :return: numerator and denominator of the approximation
    """
    assert 0 <= x
    ix = int(x)
    yield ix, 1
    if numpy.isclose(x, ix, rtol, atol):
        return
    for numer, denom in rationalizations(1.0 / (x - ix), rtol, atol):
        yield denom + ix * numer, numer


def find_fundamental(fs: list[float], rtol: float = 1e-3) -> tuple[float, npt.NDArray[numpy.int32]]:
    """
    Find the fundamental frequency of a list of frequencies

    :param fs: frequencies
    :param rtol: Relative tolerance, see numpy.isclose
    :return: fundamental frequency and multiplication factors for each frequency in fs
    """
    f0 = fs[0]
    res = numpy.zeros((len(fs), 2), dtype=numpy.int32)
    for i, f in enumerate(fs):
        res[i] = list(rationalizations(f0 / f, rtol=rtol))[-1]

    lcm = numpy.lcm.reduce(res[:, 0])
    common = (lcm // res[:, 0] * res.T)[1]
    return numpy.mean(fs / common), common
